// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
// 
//     EMail:    bochs@world.std.com



#include "bochs.h"



// boundaries of consideration:
//
//  * physical memory boundary: 1024k (1Megabyte) (increments of...)
//  * A20 boundary:             1024k (1Megabyte)
//  * page boundary:            4k
//  * ROM boundary:             2k (dont care since we are only reading)
//  * segment boundary:         any


static Bit8u *fetch_ptr = NULL;
static Bit32u bytesleft = 0;
static Bit32u linear_page;
static Bit32u max_phy_addr;


  void
bx_cpu_c::prefetch(void)
{
  // cs:eIP
  // prefetch QSIZE byte quantity aligned on corresponding boundary
  Bit32u linear_instr_addr;
  Bit32u phy_addr, max_phy_addr;
  Bit32u temp_eip, temp_limit;

  temp_eip   = BX_CPU_THIS_PTR eip;
  temp_limit = BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled;

  linear_instr_addr = BX_CPU_THIS_PTR cs.cache.u.segment.base + temp_eip;
  if (temp_eip > temp_limit) {
    bx_panic("prefetch: EIP > CS.limit\n");
    }

  if (BX_CPU_THIS_PTR cr0.pg) {
    // aligned block guarenteed to be all in one page, same A20 address
    phy_addr = itranslate_linear(linear_instr_addr, CPL==3);
    phy_addr = A20ADDR(phy_addr);
    }
  else {
    phy_addr = A20ADDR(linear_instr_addr);
    }

  if ( phy_addr >= bx_mem.len )
    return; // ??? maybe panic here

  // max physical address as confined by page boundary
  linear_page = phy_addr & 0xfffff000;
  max_phy_addr = linear_page + 0x00000fff;

  // check if segment boundary comes into play
  if ((temp_limit - temp_eip) < 4096) {
    }

  bytesleft = (max_phy_addr - phy_addr) + 1;
  fetch_ptr = &bx_mem.vector[phy_addr];
}

  // If control has transfered locally, it is possible the prefetch Q is
  // still valid.  This would happen for repeat instructions, and small
  // branches.
  void
bx_cpu_c::revalidate_prefetch_q(void)
{
  Bit32u new_linear_addr, new_linear_page, new_linear_offset;
  Bit32u new_phy_addr;

  new_linear_addr = BX_CPU_THIS_PTR cs.cache.u.segment.base + BX_CPU_THIS_PTR eip;

  new_linear_page = new_linear_addr & 0xfffff000;
  if (new_linear_page == linear_page) {
    new_linear_offset = new_linear_addr & 0x00000fff;
    new_phy_addr = new_linear_page + new_linear_offset;
    bytesleft = (max_phy_addr - new_phy_addr) + 1;
    fetch_ptr = &bx_mem.vector[new_phy_addr];
    }
  else {
    bytesleft = 0;
    }
}

  INLINE void
bx_cpu_c::invalidate_prefetch_q(void)
{
  bytesleft = 0;
}


  INLINE Bit8u
bx_cpu_c::fetch_next_byte(void)
{
  if (bytesleft == 0) {
    prefetch();
    }

  BX_INSTR_MEM_CODE(BX_CPU_THIS_PTR cs.selector.value + BX_CPU_THIS_PTR eip, 1);

  BX_CPU_THIS_PTR eip++;
  bytesleft--;
  return( *fetch_ptr++ );
}


  INLINE Bit16u
bx_cpu_c::fetch_next_word(void)
{
  Bit8u b_low, b_high;

  BX_INSTR_MEM_CODE(BX_CPU_THIS_PTR cs.selector.value + BX_CPU_THIS_PTR eip, 2);

  if (bytesleft >= 2) {
    b_low  = *fetch_ptr++;
    b_high = *fetch_ptr++;
    BX_CPU_THIS_PTR eip += 2;
    bytesleft -= 2;
    return( (b_high << 8) | b_low );
    }
  else if (bytesleft == 1) {
    b_low  = *fetch_ptr++;
    BX_CPU_THIS_PTR eip += 1;
    // bytesleft = 0;  // no need for this
    prefetch();
    b_high = *fetch_ptr++;
    BX_CPU_THIS_PTR eip += 1;
    bytesleft--;
    return( (b_high<<8) | b_low );
    }
  else {
    prefetch();
    b_low  = *fetch_ptr++;
    b_high = *fetch_ptr++;
    BX_CPU_THIS_PTR eip += 2;
    bytesleft -= 2;
    return( (b_high<<8) | b_low );
    }
}

  INLINE Bit32u
bx_cpu_c::fetch_next_dword(void)
{
  BX_INSTR_MEM_CODE(BX_CPU_THIS_PTR cs.selector.value + BX_CPU_THIS_PTR eip, 4);

  if (bytesleft >= 4) {
    Bit8u b0, b1, b2, b3;

    b0 = *fetch_ptr++;
    b1 = *fetch_ptr++;
    b2 = *fetch_ptr++;
    b3 = *fetch_ptr++;
    BX_CPU_THIS_PTR eip += 4;
    bytesleft -= 4;
    return( (b3<<24) | (b2<<16) | (b1<<8) | b0 );
    }
  else {
    Bit8u b[4];
    unsigned index;

    // code spans multiple pages
    for (index=0; index<4; index++) {
      if (bytesleft == 0)
        prefetch();
      b[index] = *fetch_ptr++;
      BX_CPU_THIS_PTR eip += 1;
      bytesleft--;
      }

    return( (b[3]<<24) | (b[2]<<16) | (b[1]<<8) | b[0] );
    }
}
