#include <stdio.h>
#include <string.h>
#include <signal.h>

#include "bochs.h"

// number of simulators we're driving
#define BX_NSIM 1

// Change these to point to the C function which
// initializes each simulation environment
#define BX_SIM1_INIT_FUNCTION    bx_bochs_init
#define BX_SIM2_INIT_FUNCTION    sim2_init

#define BX_SIM1_EXIT(code)   bx_panic("debugger exiting.\n")
#define BX_SIM2_EXIT(code)   sim2_exit(code)

// This section for declarations of functions from other simulation
// environments, which are declared in macros above.
// BEGIN ==========================================================

// Put C++ declarations here.

// Put C declarations here.
extern "C" {
  void sim2_init(int argc, char *argv[]);
  void sim2_exit(int code);
  };
// END ============================================================

#define BX_INFILE_DEPTH  10

#define INCLUDE_CMD       "source"
#define INCLUDE_CMD_LEN   strlen(INCLUDE_CMD)


void bx_dbg_usage(char *);

static char bx_debug_rc_fname[BX_MAX_PATH];

static char tmp_buf[512];
static char *tmp_buf_ptr;

static char *argv0 = NULL;
static char **sim1_argv = NULL;


typedef struct {
  FILE    *fp;
  char     fname[BX_MAX_PATH];
  unsigned lineno;
  } bx_infile_stack_entry_t;

bx_infile_stack_entry_t bx_infile_stack[BX_INFILE_DEPTH];
int                     bx_infile_stack_index = 0;

static void bx_get_command(void);
static int  bx_nest_infile(char *path);
static void bx_unnest_infile(void);

static void bx_debug_ctrlc_handler(int signum);
volatile Boolean bx_dbg_interrupt_requested = 0;


static void bx_dbg_cpu_loop(void);

static bx_bpoint_state_t bpoint_state;
bx_bpoint_requests_t bx_dbg_bpoint_requests;
static void bx_dbg_init_breakpoints(void);
static Bit8u bx_disasm_ibuf[32];
#if BX_DISASM
static char bx_disasm_tbuf[512];
#endif

// for debugging the disassembly stuff only
#define BX_DISASM_CHECK 0

#if BX_DISASM_CHECK
  void bx_dbg_disasm_check(void);
#endif

static char   default_display_format = 'x';
static char   default_unit_size      = 'w';
static Bit32u default_addr = 0;

Boolean  bx_dbg_take_dma = 1;
Boolean  bx_dbg_take_irq = 1;
Boolean  bx_dbg_async_irq = 1;
Boolean  bx_dbg_async_dma = 1;

Boolean  bx_dbg_event_reports = 0;
Boolean  bx_dbg_auto_disassemble = 0;
unsigned bx_dbg_disassemble_size = 32;

// allocate dirty page table
#define BX_MAX_DIRTY_PAGE_TABLE_MEGS 64
unsigned char bx_dbg_dirty_pages[(BX_MAX_DIRTY_PAGE_TABLE_MEGS * 1024 * 1024) / 4096];

#define BX_BATCH_DMA_BUFSIZE 512

static struct {
  unsigned this_many;  // batch this many max before posting events
  unsigned Qsize;      // this many have been batched
  struct {
    Bit32u   addr;   // address of DMA op
    unsigned len;    // number of bytes in op
    unsigned what;   // BX_READ or BX_WRITE
    Bit32u   val;    // value of DMA op
    unsigned icount; // icount at this dma op
    } Q[BX_BATCH_DMA_BUFSIZE];
  } batch_dma;

void bx_dbg_post_dma_reports(void);



  int
bx_dbg_main(int argc, char *argv[])
{
  int i, sim1_i;
  unsigned sim2_arg_index=0;
  int      sim1_argc=0, sim2_argc=0;
  int reti;

  batch_dma.this_many = 1;
  batch_dma.Qsize     = 0;

  // initialize dirty pages table
  memset(bx_dbg_dirty_pages, 0, sizeof(bx_dbg_dirty_pages));

  argv0 = strdup(argv[0]);

  sim1_argv = (char **) malloc( sizeof(char *) * argc );
  if (sim1_argv == NULL) {
    fprintf(stderr, "%s: out of memory allocating argv\n", argv[0]);
    exit(1);
    }
  sim1_argv[0] = strdup(argv[0]);

  bx_debug_rc_fname[0] = '\0';


  i = 1;
  if ( (argc >= 2) && !strcmp(argv[1], "-rc") ) {
    if ( argc == 2 ) {
      fprintf(stderr, "%s: -rc option used, but no path specified.\n",
        argv[0]);
      bx_dbg_usage(argv[0]);
      exit(1);
      }
    strncpy(bx_debug_rc_fname, argv[2], BX_MAX_PATH-1);
    i += 2; // skip past "-rc" and filename
    }


#if BX_NSIM > 1
  // if driving multiple simulators, need to split out
  // options to each one, so appropriate options can be
  // passed to each simulator
  for (; i<argc; i++) {

    if (strcmp(argv[i], "-sim1") == 0) {
      sim1_argc = 1;
      for (i=i+1; i<argc; i++) {
        if (strcmp(argv[i], "-sim2") == 0) {
          i--; // backup i
          break;
          }
        sim1_argc++;
        }
      }

    else if (strcmp(argv[i], "-sim2") == 0) {
      sim2_arg_index = i;
      sim2_argc = 1;
      for (i=i+1; i<argc; i++) {
        if (strcmp(argv[i], "-sim1") == 0) {
          i--; // backup i
          break;
          }
        sim2_argc++;
        }
      }

    else {
      fprintf(stderr, "%s: %s option not understood.\n",
        argv[0], argv[i]);
      bx_dbg_usage(argv[0]);
      exit(1);
      }
    }

#else  // BX_NSIM > 1

  // Driving single simulator.  Just pass all remaining options
  // through to simulator.
  sim1_argc = 1;
  sim1_i = 1;
  for (; i<argc; i++) {
    sim1_argv[sim1_i] = strdup(argv[i]);
    sim1_argc++;
    sim1_i++;
    }

#endif  // BX_NSIM > 1

  bx_infile_stack_index = 0;
  bx_infile_stack[0].fp = stdin;
  strncpy(bx_infile_stack[0].fname, argv[0], BX_MAX_PATH);
  bx_infile_stack[0].fname[BX_MAX_PATH-1] = 0;
  bx_infile_stack[0].lineno = 0;


  if (bx_debug_rc_fname[0] == '\0') {
    fprintf(stderr, "%s: Warning: no rc file specified.\n", argv[0]);
    }
  else {
    fprintf(stderr, "%s: using rc file '%s'.\n", argv[0], bx_debug_rc_fname);
    // if there's an error, the user will know about it before proceeding
    (void) bx_nest_infile(bx_debug_rc_fname);
    }

  bx_dbg_init_breakpoints();

  for (unsigned j=0; j<sizeof(bx_disasm_ibuf); j++)
    bx_disasm_ibuf[j] = 0;

  BX_SIM1_INIT_FUNCTION (sim1_argc, sim1_argv);
  BX_SIM2_INIT_FUNCTION (sim2_argc, &argv[sim2_arg_index]);

  if (bx_mem.megabytes > BX_MAX_DIRTY_PAGE_TABLE_MEGS) {
    fprintf(stderr, "Error: memory larger than dirty page table can handle\n");
    fprintf(stderr, "Error: increase BX_MAX_DIRTY_PAGE_TABLE_MEGS\n");
    bx_dbg_exit(1);
    }

  // setup Ctrl-C handler
  signal(SIGINT, bx_debug_ctrlc_handler);

#if BX_DISASM_CHECK
  bx_dbg_disasm_check();
#endif

  while ( 1 ) {
    bx_get_command();
    if ( (*tmp_buf_ptr == '\n') || (*tmp_buf_ptr == 0) ) {
      fprintf(stderr, "\n");
      }
    else if ( (strncmp(tmp_buf_ptr, INCLUDE_CMD, INCLUDE_CMD_LEN) == 0) &&
              (tmp_buf_ptr[INCLUDE_CMD_LEN] == ' ' ||
               tmp_buf_ptr[INCLUDE_CMD_LEN] == '\t') ) {
      char *ptr;
      int len;

      ptr = tmp_buf_ptr + INCLUDE_CMD_LEN+1;
      while ( *ptr==' ' || *ptr=='\t' )
        ptr++;

      len = strlen(ptr);
      if (len == 0) {
        fprintf(stderr, "%s: no filename given to 'source' command.\n", argv0);
        if (bx_infile_stack_index > 0) {
          fprintf(stderr, "%s: ERROR in source file causes exit.\n", argv0);
          bx_dbg_exit(1);
          }
        continue;
        }
      ptr[len-1] = 0; // get rid of newline
      fprintf(stderr, "\n");
      reti = bx_nest_infile(ptr);
      if ((reti==0) && (bx_infile_stack_index > 0)) {
        fprintf(stderr, "%s: ERROR in source file causes exit.\n", argv0);
        bx_dbg_exit(1);
        }
      }
    else {
      // Give a chance to the command line extensions, to
      // consume the command.  If they return 0, then
      // we need to process the command.  A return of 1
      // means, the extensions have handled the command
      if ( bx_dbg_extensions(tmp_buf_ptr)==0 ) {
        // process command here
        //fprintf(stderr, "read line: '%s'\n", tmp_buf_ptr);
        bx_add_lex_input(tmp_buf_ptr);
        yyparse();
        }
      }
    }

  bx_dbg_exit(0);
  return(0); // keep compiler happy
}


  void
bx_get_command(void)
{
  char *charptr_ret;

  bx_infile_stack[bx_infile_stack_index].lineno++;

  if (bx_infile_stack_index == 0) {
    fprintf(stderr, "<bochs:%d> ", bx_infile_stack[bx_infile_stack_index].lineno);
    }

  charptr_ret = fgets(tmp_buf, 512,
    bx_infile_stack[bx_infile_stack_index].fp);
  if (charptr_ret == NULL) {
    // see if error was due to EOF condition
    if (feof(bx_infile_stack[bx_infile_stack_index].fp)) {
      if (bx_infile_stack_index > 0) {
        // nested level of include files, pop back to previous one
        fprintf(stderr, "\n");
        bx_unnest_infile();
        }
      else {
        // not nested, sitting at stdin prompt, user wants out
        bx_dbg_quit();
        }

      // call recursively
      bx_get_command();
      return;
      }

    // error was not EOF, see if it was from a Ctrl-C
    if (bx_dbg_interrupt_requested) {
      tmp_buf[0] = '\n';
      tmp_buf[1] = 0;
      tmp_buf_ptr = &tmp_buf[0];
      bx_dbg_interrupt_requested = 0;
      return;
      }

    fprintf(stderr, "fgets() returned ERROR.\n");
fprintf(stderr, "intr request was %u\n", bx_dbg_interrupt_requested);
    bx_dbg_exit(1);
    }
  tmp_buf_ptr = &tmp_buf[0];

  // look for first non-whitespace character
  while ( ((*tmp_buf_ptr == ' ') || (*tmp_buf_ptr == '\t')) &&
          (*tmp_buf_ptr != '\n') && (*tmp_buf_ptr != 0) ) {
    tmp_buf_ptr++;
    }

  return;
}


  void
bx_dbg_usage(char *progname)
{
  fprintf(stderr, "usage: %s [-rc path] [-sim1 ... ] [-sim2 ... ]\n", progname);
}

  void
bx_dbg_continue(void)
{
  // continue executing, until a breakpoint or request occurs

  bx_dbg_bpoint_requests.instr_count_max = 0;  // unlimited
  bx_dbg_cpu_loop();
}

  void
bx_dbg_single_step(Bit32u count)
{
  // next instruction, stepping into function calls

  bx_dbg_bpoint_requests.instr_count_max = count;
  bx_dbg_cpu_loop();
  if (bx_dbg_event_reports)
    fprintf(stderr, "done\n");
}


  void
bx_dbg_cpu_loop(void)
{
  // zero out breakpoint request structure
  memset(&bpoint_state, 0, sizeof(bpoint_state));

  // execute simulator until a requested condition occurs
  bx_dbg_interrupt_requested = 0;
  bx_cpu.cpu_loop();
  bx_dbg_interrupt_requested = 0;

  if ( bx_dbg_bpoint_requests.take_dma || bx_dbg_bpoint_requests.take_irq ) {
    // no execution happened, only commanded CPU to take DMA or IRQ
    return;
    }

  if (bpoint_state.is_instr_count_bpoint) {
    }
  else if (bpoint_state.is_vir_iaddr_bpoint) {
    fprintf(stderr, "Breakpoint %u, 0x%x in ?? ()\n",
            bpoint_state.vir_iaddr_handle,
            bpoint_state.before.iaddr.linear_addr);
    }
  else if (bpoint_state.is_lin_iaddr_bpoint) {
    fprintf(stderr, "Breakpoint %u, 0x%x in ?? ()\n",
            bpoint_state.lin_iaddr_handle,
            bpoint_state.before.iaddr.linear_addr);
    }
  else if (bpoint_state.is_phy_iaddr_bpoint) {
    fprintf(stderr, "Breakpoint %u, 0x%x in ?? ()\n",
            bpoint_state.phy_iaddr_handle,
            bpoint_state.before.iaddr.linear_addr);
    }

  fprintf(stderr, "0x%x in ?? ()\n",
          (unsigned) bpoint_state.before.iaddr.linear_addr);

  if (bx_dbg_auto_disassemble) {
    if (bpoint_state.before.iaddr.phy_addr_valid) {
#if BX_DISASM
      unsigned ilen;

      bx_dbg_fetch_mem(bpoint_state.before.iaddr.phy_addr, 16, bx_disasm_ibuf);
      ilen = bx_disassemble.disasm(bpoint_state.before.is_32bit_code,
                                   bx_disasm_ibuf, bx_disasm_tbuf);
   
      if (bpoint_state.before.is_32bit_code) {
        fprintf(stderr, "%04x:%08x: ",
                (unsigned) bpoint_state.before.iaddr.cs,
                (unsigned) bpoint_state.before.iaddr.eip);
        }
      else {
        fprintf(stderr, "%04x:%04x: ",
                (unsigned) bpoint_state.before.iaddr.cs,
                (unsigned) bpoint_state.before.iaddr.eip);
        }
      for (unsigned j=0; j<ilen; j++)
        fprintf(stderr, "%02x", (unsigned) bx_disasm_ibuf[j]);
      fprintf(stderr, ": %s\n", bx_disasm_tbuf);
#endif  // #if BX_DISASM
      }
    else {
      fprintf(stderr, "??? (physical address not available)\n");
      }
    }

#if 0
  fprintf(stderr, "cpu_loop returned (%u).\n",
    (unsigned) bpoint_state.instr_count);
  fprintf(stderr, "  cs:eip = %04x:%08x, %04x:%08x\n",
     (unsigned) bpoint_state.before.iaddr.cs,
     (unsigned) bpoint_state.before.iaddr.eip,
     (unsigned) bpoint_state.after.iaddr.cs,
     (unsigned) bpoint_state.after.iaddr.eip);

  fprintf(stderr, "  is_instr_count_bpoint = %u\n",
    bpoint_state.is_instr_count_bpoint);
  fprintf(stderr, "  is_vir_iaddr_bpoint = %u\n",
    bpoint_state.is_vir_iaddr_bpoint);
  fprintf(stderr, "  is_lin_iaddr_bpoint = %u\n",
    bpoint_state.is_lin_iaddr_bpoint);
  fprintf(stderr, "  is_phy_iaddr_bpoint = %u\n",
    bpoint_state.is_phy_iaddr_bpoint);
  fprintf(stderr, "  is_io_bpoint = %u\n",
    bpoint_state.is_io_bpoint);
  fprintf(stderr, "  is_soft_int_bpoint = %u\n",
    bpoint_state.is_soft_int_bpoint);
  fprintf(stderr, "  is_hard_int_bpoint = %u\n",
    bpoint_state.is_hard_int_bpoint);
  fprintf(stderr, "  is_memory_bpoint = %u\n",
    bpoint_state.is_memory_bpoint);
  fprintf(stderr, "  is_opcode_bpoint = %u\n",
    bpoint_state.is_opcode_bpoint);
  fprintf(stderr, "  is_exception_bpoint = %u\n",
    bpoint_state.is_exception_bpoint);
  fprintf(stderr, "  is_user_interrupt = %u\n",
    bpoint_state.is_user_interrupt);
#endif
}


  void
bx_dbg_set(char *p1, char *p2, char *p3)
{
  fprintf(stderr, "Error: %s %s %s: command 'set' not implemented yet.\n",
    p1, p2, p3);
}

  void
bx_dbg_breakpoint(char *p1, char *p2)
{
  fprintf(stderr, "%s %s: command 'b' not implemented yet.\n",
    p1, p2);
}

  void
bx_dbg_del_breakpoint(unsigned handle)
{
  if (handle >= bx_dbg_bpoint_requests.num_iaddr_bpoints) {
    fprintf(stderr, "No breakpoint number %u.\n", handle);
    return;
    }

  bx_dbg_bpoint_requests.bpoint_type[handle] = BX_BPOINT_TYPE_NONE;
}

  int
bx_nest_infile(char *path)
{
  FILE *tmp_fp;
 
  tmp_fp = fopen(path, "r");
  if (!tmp_fp) {
    fprintf(stderr, "%s: can not open file '%s' for reading.\n",
            argv0, path);
    return(0);
    }
 
  if ( (bx_infile_stack_index+1) >= BX_INFILE_DEPTH ) {
    fprintf(stderr, "%s: source files nested too deeply\n", argv0);
    return(0);
    }
 
  bx_infile_stack_index++;
  bx_infile_stack[bx_infile_stack_index].fp = tmp_fp;
  strncpy(bx_infile_stack[bx_infile_stack_index].fname, path, BX_MAX_PATH);
  bx_infile_stack[bx_infile_stack_index].fname[BX_MAX_PATH-1] = 0;
  bx_infile_stack[bx_infile_stack_index].lineno = 0;
  return(1);
}

  void
bx_unnest_infile(void)
{
  if (bx_infile_stack_index <= 0) {
    fprintf(stderr, "%s: ERROR: unnest_infile(): nesting level = 0.\n",
      argv0);
    bx_dbg_exit(1);
    }

  fclose(bx_infile_stack[bx_infile_stack_index].fp);
  bx_infile_stack_index--;
}

  int
yywrap(void)
{
  fprintf(stderr, "%s: ERROR: yywrap() called.\n", argv0);
  bx_dbg_exit(1);
  return(0); // keep compiler quiet
}


  void
yyerror(char *s)
{
  fprintf(stderr, "%s:%d: %s at '%s'\n",
    bx_infile_stack[bx_infile_stack_index].fname,
    bx_infile_stack[bx_infile_stack_index].lineno,
    s, yytext);

  if (bx_infile_stack_index > 0) {
    fprintf(stderr, "%s: ERROR in source file causes exit.\n", argv0);
    bx_dbg_exit(1);
    }
}

  void
bx_debug_ctrlc_handler(int signum)
{
  UNUSED(signum);

  signal(SIGINT, bx_debug_ctrlc_handler);
  bx_dbg_interrupt_requested = 1;
}

  void
bx_dbg_exit(int code)
{
fprintf(stderr, "before sim1_exit\n");
 bx_panic("debugger exiting.\n");
  //BX_SIM1_EXIT(code);
fprintf(stderr, "before sim2_exit\n");
  BX_SIM2_EXIT(code);
fprintf(stderr, "before exit\n");

  exit(code);
}

  void
bx_dbg_vbreakpoint(Boolean specific, Bit32u cs, Bit32u eip)
{
  unsigned int i = bx_dbg_bpoint_requests.num_iaddr_bpoints;

  if (cs > 0xffff) {
    fprintf(stderr, "%s: virtual breakpoint: CS greater than 0xFFFF.\n", argv0);
    return;
    }

  if (specific) {
    // if no more slots left, use a deleted slot if any
    if ( (i+1) > BX_MAX_IADDR_BPOINTS) {
      for (i=0; i<BX_MAX_IADDR_BPOINTS; i++) {
        if (bx_dbg_bpoint_requests.bpoint_type[i] == BX_BPOINT_TYPE_NONE)
          break;
        }
      if (i >= BX_MAX_IADDR_BPOINTS) {
        fprintf(stderr, "%s: virtual breakpoint: no bpoint slots left.\n", argv0);
        return;
        }
      }
    else
      bx_dbg_bpoint_requests.num_iaddr_bpoints++;

    bx_dbg_bpoint_requests.bpoint_type[i] = BX_BPOINT_TYPE_VIRTUAL;
    bx_dbg_bpoint_requests.iaddr[i].vir.cs  = cs;
    bx_dbg_bpoint_requests.iaddr[i].vir.eip = eip;
    bx_dbg_bpoint_requests.attr[i].disposition = bx_disp_keep;
    bx_dbg_bpoint_requests.attr[i].enabled = 1;
    bx_dbg_bpoint_requests.attr[i].ignorecount = 0;
    }
  else
    fprintf(stderr, "VBREAK not implemented (not-specific)\n");
}


  void
bx_dbg_lbreakpoint(Boolean specific, Bit32u laddress)
{
  unsigned int i = bx_dbg_bpoint_requests.num_iaddr_bpoints;

  if (specific) {
    // if no more slots left, use a deleted slot if any
    if ( (i+1) > BX_MAX_IADDR_BPOINTS) {
      for (i=0; i<BX_MAX_IADDR_BPOINTS; i++) {
        if (bx_dbg_bpoint_requests.bpoint_type[i] == BX_BPOINT_TYPE_NONE)
          break;
        }
      if (i >= BX_MAX_IADDR_BPOINTS) {
        fprintf(stderr, "%s: linear breakpoint: no bpoint slots left.\n", argv0);
        return;
        }
      }
    else
      bx_dbg_bpoint_requests.num_iaddr_bpoints++;

    bx_dbg_bpoint_requests.bpoint_type[i] = BX_BPOINT_TYPE_LINEAR;
    bx_dbg_bpoint_requests.iaddr[i].lin.addr = laddress;
    bx_dbg_bpoint_requests.attr[i].disposition = bx_disp_keep;
    bx_dbg_bpoint_requests.attr[i].enabled = 1;
    bx_dbg_bpoint_requests.attr[i].ignorecount = 0;
    }
  else
    fprintf(stderr, "LBREAK not implemented (not-specific)\n");
}

  void
bx_dbg_pbreakpoint(Boolean specific, Bit32u paddress)
{
  unsigned int i = bx_dbg_bpoint_requests.num_iaddr_bpoints;

  if (specific) {
    // if no more slots left, use a deleted slot if any
    if ( (i+1) > BX_MAX_IADDR_BPOINTS) {
      for (i=0; i<BX_MAX_IADDR_BPOINTS; i++) {
        if (bx_dbg_bpoint_requests.bpoint_type[i] == BX_BPOINT_TYPE_NONE)
          break;
        }
      if (i >= BX_MAX_IADDR_BPOINTS) {
        fprintf(stderr, "%s: physical breakpoint: no bpoint slots left.\n", argv0);
        return;
        }
      }
    else
      bx_dbg_bpoint_requests.num_iaddr_bpoints++;

    bx_dbg_bpoint_requests.bpoint_type[i] = BX_BPOINT_TYPE_PHYSICAL;
    bx_dbg_bpoint_requests.iaddr[i].phy.addr = paddress;
    bx_dbg_bpoint_requests.attr[i].disposition = bx_disp_keep;
    bx_dbg_bpoint_requests.attr[i].enabled = 1;
    bx_dbg_bpoint_requests.attr[i].ignorecount = 0;
    }
  else
    fprintf(stderr, "PBREAK not implemented (not-specific)\n");
}


  void
bx_dbg_init_breakpoints(void)
{
  unsigned i;
  bx_dbg_bpoint_requests.instr_count_max = 0;

  bx_dbg_bpoint_requests.num_iaddr_bpoints = 0;
  for (i=0; i<BX_MAX_IADDR_BPOINTS; i++) {
    bx_dbg_bpoint_requests.bpoint_type[i] = BX_BPOINT_TYPE_NONE;
    }
}


  Boolean
bx_dbg_cpu_icount(Bit32u cs, Bit32u eip,
                  Bit32u laddr, Bit32u paddr, Boolean paddr_valid,
                  Boolean intr_pending, Boolean is_32bit_code)
{
  // function returns
  //   0: CPU should continue executing
  //   1: CPU should return immediately to debug code

  if ( bx_dbg_bpoint_requests.take_irq ||
       bx_dbg_bpoint_requests.take_dma ) {
    return(1);
    }

  // if any break condition has occurred DURING execution of
  // last instruction, store state info about next instruction,
  // and tell CPU to return to the debug environment
  if (bpoint_state.event_occurred) {
    bpoint_state.after.intr_pending = intr_pending;
    bpoint_state.after.iaddr.cs = cs;
    bpoint_state.after.iaddr.eip = eip;
    bpoint_state.after.iaddr.linear_addr = laddr;
    bpoint_state.after.iaddr.phy_addr = paddr;
    bpoint_state.after.iaddr.phy_addr_valid = paddr_valid;
    bpoint_state.after.is_32bit_code = is_32bit_code;
    return(1);
    }

  // store state info about current instruction to be executed next
  bpoint_state.before.intr_pending = intr_pending;
  bpoint_state.before.iaddr.cs = cs;
  bpoint_state.before.iaddr.eip = eip;
  bpoint_state.before.iaddr.linear_addr = laddr;
  bpoint_state.before.iaddr.phy_addr = paddr;
  bpoint_state.before.iaddr.phy_addr_valid = paddr_valid;
  bpoint_state.before.is_32bit_code = is_32bit_code;

  //
  // Now look for break conditions BEFORE execution of next instruction
  //

  // look for Ctrl-C
  if (bx_dbg_interrupt_requested) {
    bpoint_state.is_user_interrupt = 1;
    bpoint_state.event_occurred = 1;
    }

  for (unsigned i=0; i<bx_dbg_bpoint_requests.num_iaddr_bpoints; i++) {

    if ( bx_dbg_bpoint_requests.bpoint_type[i] == BX_BPOINT_TYPE_VIRTUAL ) {
      // look for breakpoints at this instruction's virtual address
      if ( (bx_dbg_bpoint_requests.iaddr[i].vir.cs  == cs) &&
           (bx_dbg_bpoint_requests.iaddr[i].vir.eip == eip) ) {
        if (bpoint_state.instr_count==0) continue;
        bpoint_state.event_occurred = 1;
        bpoint_state.is_vir_iaddr_bpoint = 1;
        bpoint_state.vir_iaddr_handle = i;
        }
      }

    else if ( bx_dbg_bpoint_requests.bpoint_type[i] == BX_BPOINT_TYPE_LINEAR ) {
      // look for breakpoints at this instruction's linear address
      if ( bx_dbg_bpoint_requests.iaddr[i].lin.addr  == laddr ) {
        if (bpoint_state.instr_count==0) continue;
        bpoint_state.event_occurred = 1;
        bpoint_state.is_lin_iaddr_bpoint = 1;
        bpoint_state.lin_iaddr_handle = i;
        }
      }

    else if ( bx_dbg_bpoint_requests.bpoint_type[i] == BX_BPOINT_TYPE_PHYSICAL ) {
      // look for breakpoints at this instruction's physical address
      if ( bx_dbg_bpoint_requests.iaddr[i].phy.addr  == paddr ) {
        if (bpoint_state.instr_count==0) continue;
        bpoint_state.event_occurred = 1;
        bpoint_state.is_phy_iaddr_bpoint = 1;
        bpoint_state.phy_iaddr_handle = i;
        }
      }
    }


  // Look for a step count
  if (bx_dbg_bpoint_requests.instr_count_max &&
      ((bpoint_state.instr_count+1) > bx_dbg_bpoint_requests.instr_count_max)) {
    bpoint_state.is_instr_count_bpoint = 1;
    bpoint_state.event_occurred = 1;
    }

  if (bpoint_state.event_occurred) return(1);

  bpoint_state.instr_count++;
  return(0);
}

  void
bx_dbg_info_bpoints(void)
{
// Num Type           Disp Enb Address    What
// 1   breakpoint     keep y   0x00010664 in main at temp.c:7

  fprintf(stderr, "Num Type           Disp Enb Address\n");
  for (unsigned i=0; i<bx_dbg_bpoint_requests.num_iaddr_bpoints; i++) {
    if (bx_dbg_bpoint_requests.bpoint_type[i] != BX_BPOINT_TYPE_NONE) {
      fprintf(stderr, "%3u ", i);
      switch (bx_dbg_bpoint_requests.bpoint_type[i]) {
        case BX_BPOINT_TYPE_VIRTUAL:  fprintf(stderr, "vbreakpoint    "); break;
        case BX_BPOINT_TYPE_LINEAR:   fprintf(stderr, "lbreakpoint    "); break;
        case BX_BPOINT_TYPE_PHYSICAL: fprintf(stderr, "pbreakpoint    "); break;
        }
      switch (bx_dbg_bpoint_requests.attr[i].disposition) {
        case bx_disp_disable: fprintf(stderr, "dis  "); break;
        case bx_disp_keep:    fprintf(stderr, "keep "); break;
        case bx_disp_delete:  fprintf(stderr, "del  "); break;
        }
      if (bx_dbg_bpoint_requests.attr[i].enabled)
        fprintf(stderr, "y   ");
      else
        fprintf(stderr, "n   ");

      switch (bx_dbg_bpoint_requests.bpoint_type[i]) {
        case BX_BPOINT_TYPE_VIRTUAL:
          fprintf(stderr, "0x%04x:0x%08x\n",
                  bx_dbg_bpoint_requests.iaddr[i].vir.cs,
                  bx_dbg_bpoint_requests.iaddr[i].vir.eip);
          break;
        case BX_BPOINT_TYPE_LINEAR:
          fprintf(stderr, "0x%08x\n",
                  bx_dbg_bpoint_requests.iaddr[i].lin.addr);
          break;
        case BX_BPOINT_TYPE_PHYSICAL:
          fprintf(stderr, "0x%08x\n",
                  bx_dbg_bpoint_requests.iaddr[i].phy.addr);
          break;
        }
      }
    }
}

  void
bx_dbg_quit(void)
{
  fprintf(stderr, "Quit\n");
  bx_dbg_exit(0);
}


#if BX_DISASM_CHECK
  void
bx_dbg_disasm_check(void)
{
  unsigned ilen;
  unsigned or16=0, or32=0; // use segment override in 16 & 32bit modes (0 or 1)

  fprintf(stderr, "\nDISASM CHECK: 16bit address mode\n");

  bx_disasm_ibuf[0] = 0x26;      // ES: (if or16==0, not used)
  bx_disasm_ibuf[0+or16] = 0x8a; // mov_gbeb AL, ...
  bx_disasm_ibuf[2+or16] = 0x11; // 8bit address data if needed
  bx_disasm_ibuf[3+or16] = 0x22; // 16bit address data if needed

  for (unsigned mod=0; mod<=3; mod++) {
    for (unsigned rm=0; rm<=7; rm++) {
      bx_disasm_ibuf[1+or16] = (mod << 6) | rm;
      ilen = bx_disassemble.disasm(0, bx_disasm_ibuf, bx_disasm_tbuf);
 
      for (unsigned j=0; j<ilen; j++)
        fprintf(stderr, "%02x", (unsigned) bx_disasm_ibuf[j]);
      fprintf(stderr, ": %s\n", bx_disasm_tbuf);
      }
    }

  fprintf(stderr, "\nDISASM CHECK: 32bit address mode\n");
  fprintf(stderr, "\nOne byte opcodes:\n");

  bx_disasm_ibuf[0] = 0x26;      // ES: (if or32==0, not used)
  bx_disasm_ibuf[0+or32] = 0x67; // ADDRSIZE
  bx_disasm_ibuf[1+or32] = 0x8a; // mov_gbeb AL, ...
  bx_disasm_ibuf[3+or32] = 0x11; // address data bits  0.. 7
  bx_disasm_ibuf[4+or32] = 0x22; // address data bits  8..15
  bx_disasm_ibuf[5+or32] = 0x33; // address data bits 16..23
  bx_disasm_ibuf[6+or32] = 0x44; // address data bits 24..31


  for (unsigned mod=0; mod<=3; mod++) {
    for (unsigned rm=0; rm<=7; rm++) {
      if ( (rm==4) && (mod!=3) ) continue;
      bx_disasm_ibuf[2+or32] = (mod << 6) | rm;
      ilen = bx_disassemble.disasm(0, bx_disasm_ibuf, bx_disasm_tbuf);
 
      for (unsigned j=0; j<ilen; j++)
        fprintf(stderr, "%02x", (unsigned) bx_disasm_ibuf[j]);
      fprintf(stderr, ": %s\n", bx_disasm_tbuf);
      }
    }

  fprintf(stderr, "\nTwo byte opcodes:\n");

  bx_disasm_ibuf[0] = 0x26;      // ES: (if or32==0, not used)
  bx_disasm_ibuf[0+or32] = 0x67; // ADDRSIZE
  bx_disasm_ibuf[1+or32] = 0x8a; // mov_gbeb AL, ...
  bx_disasm_ibuf[4+or32] = 0x11; // address data bits  0.. 7
  bx_disasm_ibuf[5+or32] = 0x22; // address data bits  8..15
  bx_disasm_ibuf[6+or32] = 0x33; // address data bits 16..23
  bx_disasm_ibuf[7+or32] = 0x44; // address data bits 24..31

  for (unsigned mod=0; mod<=2; mod++) {
    unsigned rm, ss, index;

    rm = 4;
    bx_disasm_ibuf[2+or32] = (mod << 6) | rm;
    for (unsigned base=0; base<=7; base++) {
      ss = 3;
      index = 1;
      bx_disasm_ibuf[3+or32] = (ss << 6) | (index<<3) | base;
      ilen = bx_disassemble.disasm(0, bx_disasm_ibuf, bx_disasm_tbuf);
 
      for (unsigned j=0; j<ilen; j++)
        fprintf(stderr, "%02x", (unsigned) bx_disasm_ibuf[j]);
      fprintf(stderr, ": %s\n", bx_disasm_tbuf);
      }
    }
}
#endif



  void
bx_dbg_info_registers(void)
{
  Bit32u reg;

  reg = bx_dbg_get_register(BX_DBG_REG_EAX);
  fprintf(stderr, "eax            0x%-8x\t%d\n", (unsigned) reg, (int) reg);
  reg = bx_dbg_get_register(BX_DBG_REG_ECX);
  fprintf(stderr, "ecx            0x%-8x\t%d\n", (unsigned) reg, (int) reg);
  reg = bx_dbg_get_register(BX_DBG_REG_EDX);
  fprintf(stderr, "edx            0x%-8x\t%d\n", (unsigned) reg, (int) reg);
  reg = bx_dbg_get_register(BX_DBG_REG_EBX);
  fprintf(stderr, "ebx            0x%-8x\t%d\n", (unsigned) reg, (int) reg);

  reg = bx_dbg_get_register(BX_DBG_REG_ESP);
  fprintf(stderr, "esp            0x%-8x\t0x%-8x\n", (unsigned) reg, (int) reg);
  reg = bx_dbg_get_register(BX_DBG_REG_EBP);
  fprintf(stderr, "ebp            0x%-8x\t0x%-8x\n", (unsigned) reg, (int) reg);
  reg = bx_dbg_get_register(BX_DBG_REG_ESI);
  fprintf(stderr, "esi            0x%-8x\t%d\n", (unsigned) reg, (int) reg);
  reg = bx_dbg_get_register(BX_DBG_REG_EDI);
  fprintf(stderr, "edi            0x%-8x\t%d\n", (unsigned) reg, (int) reg);

  reg = bx_dbg_get_register(BX_DBG_REG_EIP);
  fprintf(stderr, "eip            0x%-8x\t0x%-8x\n", (unsigned) reg, (int) reg);

  reg = bx_dbg_get_register(BX_DBG_REG_EFLAGS);
  fprintf(stderr, "eflags         0x%-8x\t%d\n", (unsigned) reg, (int) reg);

  reg = bx_dbg_get_register(BX_DBG_REG_CS);
  fprintf(stderr, "cs             0x%-8x\t%d\n", (unsigned) reg, (int) reg);
  reg = bx_dbg_get_register(BX_DBG_REG_SS);
  fprintf(stderr, "ss             0x%-8x\t%d\n", (unsigned) reg, (int) reg);
  reg = bx_dbg_get_register(BX_DBG_REG_DS);
  fprintf(stderr, "ds             0x%-8x\t%d\n", (unsigned) reg, (int) reg);
  reg = bx_dbg_get_register(BX_DBG_REG_ES);
  fprintf(stderr, "es             0x%-8x\t%d\n", (unsigned) reg, (int) reg);
  reg = bx_dbg_get_register(BX_DBG_REG_FS);
  fprintf(stderr, "fs             0x%-8x\t%d\n", (unsigned) reg, (int) reg);
  reg = bx_dbg_get_register(BX_DBG_REG_GS);
  fprintf(stderr, "gs             0x%-8x\t%d\n", (unsigned) reg, (int) reg);
}

  void
bx_dbg_info_program(void)
{
  fprintf(stderr, "        Using the running image of child process -1.\n");
  fprintf(stderr, "Program stopped at 0x0.\n");
  fprintf(stderr, "It stopped at breakpoint 0.\n");
}

  void
bx_dbg_examine(char *command, char *format, Boolean format_passed,
               Bit32u addr, Boolean addr_passed)
{
  unsigned repeat_count, i;
  char ch, display_format, unit_size;
  Boolean iteration;
  unsigned data_size;
  Boolean paddr_valid;
  Bit32u  paddr;
  Bit8u   data8;
  Bit16u  data16;
  Bit32u  data32;
  unsigned columns, per_line, offset;
  unsigned char digit;
  unsigned biti;
  Boolean is_linear;
  unsigned char databuf[8];

  // If command was the extended "xp" command, meaning eXamine Physical memory,
  // then flag memory address as physical, rather than linear.
  if (strcmp(command, "xp") == 0) {
    is_linear = 0;
    }
  else {
    is_linear = 1;
    }

  if (addr_passed==0)
    addr = default_addr;

  if (format_passed==0) {
    display_format = default_display_format;
    unit_size      = default_unit_size;
    repeat_count   = 1;
    }
  else {
    if (format==NULL) {
      fprintf(stderr, "dbg_examine: format NULL\n");
      bx_dbg_exit(1);
      }

    if (strlen(format) < 2) {
      fprintf(stderr, "dbg_examine: invalid format passed.\n");
      bx_dbg_exit(1);
      }

    if (format[0] != '/') {
      fprintf(stderr, "dbg_examine: '/' is not first char of format.\n");
      bx_dbg_exit(1);
      }

    format++;
    repeat_count = 0;
    ch = *format;
    iteration = 0;

    while ( (ch>='0') && (ch<='9') ) {
      iteration = 1;
      repeat_count = 10*repeat_count + (ch-'0');
      format++;
      ch = *format;
      }

    if (iteration==0) {
      // if no count given, use default
      repeat_count = 1;
      }
    else if (repeat_count==0) {
      // count give, but zero is an error
      fprintf(stderr, "dbg_examine: repeat count given but is zero.\n");
      return;
      }


    // set up the default display format and unit size parameters
    display_format = default_display_format;
    unit_size      = default_unit_size;

    for (i=0; i<=1; i++) {
      if (ch==0) break; // bail on null character
      switch (ch) {
        case 'x': // hex
        case 'd': // signed decimal
        case 'u': // unsigned decimal
        case 'o': // octal
        case 't': // binary
        case 's': // null terminated string
        case 'i': // machine instruction
          display_format = ch;
          break;

        case 'b': // bytes
        case 'h': // halfwords (two bytes)
        case 'w': // words (4 bytes)
        case 'g': // giant words (8 bytes)
          unit_size = ch;
          break;
        default:
          fprintf(stderr, "dbg_examine: invalid format passed.\n");
          bx_dbg_exit(1);
          break;
        }
      format++;
      ch = *format;
      }

    // store current options as default
    default_display_format = display_format;
    default_unit_size      = unit_size;
    }

  //fprintf(stderr, "  repeat count was %u\n", repeat_count);
  //fprintf(stderr, "  display_format = '%c'\n", display_format);
  //fprintf(stderr, "  unit_size      = '%c'\n", unit_size);

  if ( (display_format == 'i') || (display_format == 's') ) {
    fprintf(stderr, "error: dbg_examine: 'i' and 's' formats not supported.\n");
    return;
    }

  if (unit_size == 'g') {
    fprintf(stderr, "error: dbg_examine: 'g' (8-byte) unit size not supported.\n");
    return;
    }

  data_size = 0;
  per_line  = 0;
  offset = 0;

  switch (unit_size) {
    case 'b': data_size = 1; per_line = 8; break;
    case 'h': data_size = 2; per_line = 8; break;
    case 'w': data_size = 4; per_line = 4; break;
    //case 'g': data_size = 8; per_line = 2; break;
    }

  columns = per_line + 1; // set current number columns past limit

  for (i=1; i<=repeat_count; i++) {

    if (columns > per_line) {
      // if not 1st run, need a newline from last line
      if (i!=1)
        fprintf(stderr, "\n");
      fprintf(stderr, "0x%x <bogus+%u>:", addr, offset);
      columns = 1;
      }

    if (is_linear) {
      dbg_xlate_linear2phy(addr, &paddr, &paddr_valid);
      if (!paddr_valid) {
        fprintf(stderr, "error: examine memory: no tranlation for linear-to-phy mem available.\n");
        return;
        }
      }
    else {
      paddr = addr;  // address is already physical address
      }

    bx_dbg_fetch_mem(paddr, data_size, databuf);
    switch (data_size) {
      case 1:
        data8 = databuf[0];
        switch (display_format) {
          case 'x': fprintf(stderr, "\t0x%02x", (unsigned) data8); break;
          case 'd': fprintf(stderr, "\t%d", (int) (Bit8s) data8); break;
          case 'u': fprintf(stderr, "\t%u", (unsigned) data8); break;
          case 'o': fprintf(stderr, "\t%o", (unsigned) data8); break;
          case 't':
            fputc('\t', stderr);
            for (biti=7; ; biti--) {
              digit = (data8 >> biti) & 0x01;
              fputc(digit + '0', stderr);
              if (biti==0) break;
              }
            break;
          }
        break;

      case 2:
#ifdef BX_LITTLE_ENDIAN
        data16 = * (Bit16u *) databuf;
#else
        data16 = (databuf[1]<<8)  |  databuf[0];
#endif
        switch (display_format) {
          case 'x': fprintf(stderr, "\t0x%04x", (unsigned) data16); break;
          case 'd': fprintf(stderr, "\t%d", (int) (Bit16s) data16); break;
          case 'u': fprintf(stderr, "\t%u", (unsigned) data16); break;
          case 'o': fprintf(stderr, "\t%o", (unsigned) data16); break;
          case 't':
            fputc('\t', stderr);
            for (biti=15; ; biti--) {
              digit = (data16 >> biti) & 0x01;
              fputc(digit + '0', stderr);
              if (biti==0) break;
              }
            break;
          }
        break;

      case 4:
#ifdef BX_LITTLE_ENDIAN
        data32 = * (Bit32u *) databuf;
#else
        data32 = (databuf[3]<<24) | (databuf[2]<<16) |
                 (databuf[1]<<8)  |  databuf[0];
#endif
        switch (display_format) {
          case 'x': fprintf(stderr, "\t0x%08x", (unsigned) data32); break;
          case 'd': fprintf(stderr, "\t%d", (int) (Bit32s) data32); break;
          case 'u': fprintf(stderr, "\t%u", (unsigned) data32); break;
          case 'o': fprintf(stderr, "\t%o", (unsigned) data32); break;
          case 't':
            fputc('\t', stderr);
            for (biti=31; ; biti--) {
              digit = (data32 >> biti) & 0x01;
              fputc(digit + '0', stderr);
              if (biti==0) break;
              }
            break;
          }
        break;
      }

    addr += data_size;
    default_addr = addr;
    columns++;
    offset += data_size;
    }
  fprintf(stderr, "\n");
}

  void
bx_dbg_setpmem(Bit32u addr, unsigned len, Bit32u val)
{
  Boolean is_OK;
  Bit8u   buf[4];

  switch ( len ) {
    case 1:
      buf[0] = (Bit8u) val;
      break;
    case 2:
      buf[0] = val & 0xff;
      buf[1] = (val>>8) & 0xff;
      break;
    case 4:
      buf[0] = val & 0xff; val >>= 8;
      buf[1] = val & 0xff; val >>= 8;
      buf[2] = val & 0xff; val >>= 8;
      buf[3] = val & 0xff;
      break;
    default:
      fprintf(stderr, "Error: setpmem: bad length value = %u\n", len);
      return;
    }

  is_OK = bx_dbg_set_mem(addr, len, buf);
  if (!is_OK) {
    fprintf(stderr, "Error: setpmem: could not set memory, out of physical bounds?\n");
    }
}

  void
bx_dbg_set_symbol(char *symbol, Bit32u val)
{
  Boolean is_OK;
  symbol++; // get past '$'

  if ( !strcmp(symbol, "eax") ) {
    is_OK = bx_dbg_set_reg(BX_DBG_REG_EAX, val);
    }
  else if ( !strcmp(symbol, "ecx") ) {
    is_OK = bx_dbg_set_reg(BX_DBG_REG_ECX, val);
    }
  else if ( !strcmp(symbol, "edx") ) {
    is_OK = bx_dbg_set_reg(BX_DBG_REG_EDX, val);
    }
  else if ( !strcmp(symbol, "ebx") ) {
    is_OK = bx_dbg_set_reg(BX_DBG_REG_EBX, val);
    }
  else if ( !strcmp(symbol, "esp") ) {
    is_OK = bx_dbg_set_reg(BX_DBG_REG_ESP, val);
    }
  else if ( !strcmp(symbol, "ebp") ) {
    is_OK = bx_dbg_set_reg(BX_DBG_REG_EBP, val);
    }
  else if ( !strcmp(symbol, "esi") ) {
    is_OK = bx_dbg_set_reg(BX_DBG_REG_ESI, val);
    }
  else if ( !strcmp(symbol, "edi") ) {
    is_OK = bx_dbg_set_reg(BX_DBG_REG_EDI, val);
    }
  else if ( !strcmp(symbol, "eip") ) {
    is_OK = bx_dbg_set_reg(BX_DBG_REG_EIP, val);
    }
  else if ( !strcmp(symbol, "eflags") ) {
    is_OK = bx_dbg_set_reg(BX_DBG_REG_EFLAGS, val);
    }
  else if ( !strcmp(symbol, "cs") ) {
    is_OK = bx_dbg_set_reg(BX_DBG_REG_CS, val);
    }
  else if ( !strcmp(symbol, "ss") ) {
    is_OK = bx_dbg_set_reg(BX_DBG_REG_SS, val);
    }
  else if ( !strcmp(symbol, "ds") ) {
    is_OK = bx_dbg_set_reg(BX_DBG_REG_DS, val);
    }
  else if ( !strcmp(symbol, "es") ) {
    is_OK = bx_dbg_set_reg(BX_DBG_REG_ES, val);
    }
  else if ( !strcmp(symbol, "fs") ) {
    is_OK = bx_dbg_set_reg(BX_DBG_REG_FS, val);
    }
  else if ( !strcmp(symbol, "gs") ) {
    is_OK = bx_dbg_set_reg(BX_DBG_REG_GS, val);
    }
  else if ( !strcmp(symbol, "synchronous_dma") ) {
    bx_dbg_take_dma = !val;  // current setting
    bx_dbg_async_dma = !val; // default setting
    return;
    }
  else if ( !strcmp(symbol, "synchronous_irq") ) {
    bx_dbg_take_irq = !val;  // current setting
    bx_dbg_async_irq = !val; // default setting
    return;
    }
  else if ( !strcmp(symbol, "event_reports") ) {
    bx_dbg_event_reports = val;
    return;
    }
  else if ( !strcmp(symbol, "auto_disassemble") ) {
    bx_dbg_auto_disassemble = (val > 0);
    return;
    }
  else if ( !strcmp(symbol, "disassemble_size") ) {
    if ( (val!=16) && (val!=32) ) {
      fprintf(stderr, "Error: disassemble_size must be 16 or 32.\n");
      return;
      }
    bx_dbg_disassemble_size = val;
    return;
    }
  else {
    fprintf(stderr, "Error: set: unrecognized symbol.\n");
    return;
    }

  if (!is_OK) {
    fprintf(stderr, "Error: could not set register '%s'.\n", symbol);
    }
}

  void
bx_dbg_query(char *what)
{
  if ( !strcmp(what, "pending") ) {
    bx_dbg_query_pending();
    }
  else {
    fprintf(stderr, "Error: Query '%s' not understood.\n", what);
    }
}

  void
bx_dbg_take(char *what, unsigned n)
{
  if ( !strcmp(what, "dma") ) {
    unsigned i;
    if (n == 0) {
      fprintf(stderr, "Error: take what n=0.\n");
      return;
      }
    bx_dbg_post_dma_reports(); // in case there's some pending reports
    batch_dma.this_many = n;

    bx_dbg_bpoint_requests.take_dma = 1;
    bx_dbg_take_dma = 1; // CPU uses this flag, force to take
    for (i=0; i<n; i++) {
      bx_dbg_cpu_loop();
      }
    bx_dbg_bpoint_requests.take_dma = 0; // zero out request
    bx_dbg_take_dma = bx_dbg_async_dma; // reset to normal mode

    batch_dma.this_many = 1;  // reset to normal
    bx_dbg_post_dma_reports(); // print reports and flush
    if (bx_dbg_event_reports)
      fprintf(stderr, "done\n");
    }
  else if ( !strcmp(what, "irq") ) {
    bx_dbg_bpoint_requests.take_irq = 1;
    bx_dbg_take_irq = 1; // CPU uses this flag, force to take
    bx_dbg_cpu_loop();
    bx_dbg_bpoint_requests.take_irq = 0; // zero out request
    bx_dbg_take_irq = bx_dbg_async_irq; // reset to normal mode
    if (bx_dbg_event_reports)
      fprintf(stderr, "done\n");
    }
  else {
    fprintf(stderr, "Error: Take '%s' not understood.\n", what);
    }
}

  void
bx_dbg_post_dma_reports(void)
{
  unsigned i;
  unsigned addr, len, what, val;
  unsigned last_addr, last_len, last_what;
  unsigned print_header;
  unsigned first_iteration;

  if (bx_dbg_event_reports) {
    if (batch_dma.Qsize == 0) return; // nothing batched to print

    // compress output so all contiguous DMA ops of the same type and size
    // are printed on the same line
    last_addr = batch_dma.Q[0].addr;
    last_len  = batch_dma.Q[0].len;
    last_what = batch_dma.Q[0].what;
    first_iteration = 1;

    for (i=0; i<batch_dma.Qsize; i++) {
      addr = batch_dma.Q[i].addr;
      len  = batch_dma.Q[i].len;
      what = batch_dma.Q[i].what;
      val  = batch_dma.Q[i].val;

      if (len != last_len)
        print_header = 1;
      else if (what != last_what)
        print_header = 1;
      else if (addr != (last_addr + last_len))
        print_header = 1;
      else 
        print_header = 0;

      // now store current values for next iteration
      last_addr = addr;
      last_len  = len;
      last_what = what;

      if (print_header) {
        if (!first_iteration) // need return from previous line
          fprintf(stderr, "\n");
        else
          first_iteration = 0;
        // need to output the event header
        fprintf(stderr, "event icount=%u DMA addr=0x%x size=%u op=%s val=0x%x",
                         batch_dma.Q[i].icount,
                         addr, len, (what==BX_READ) ? "read" : "write",
                         val );
        print_header = 0;
        }
      else {
        // *no* need to output the event header
        fprintf(stderr, " 0x%x", val);
        }
      }
    if (batch_dma.Qsize)
      fprintf(stderr, "\n");
    }

  // empty Q, regardless of whether reports are printed
  batch_dma.Qsize = 0;
}



  void
bx_dbg_dma_report(Bit32u addr, unsigned len, unsigned what, Bit32u val)
{
  if (batch_dma.this_many == 0) {
    fprintf(stderr, "%s: DMA batch this_many=0.\n", argv0);
    bx_dbg_exit(1);
    }

  // if Q is full, post events (and flush)
  if (batch_dma.Qsize >= batch_dma.this_many) {
    fprintf(stderr, "%s: DMA batch Q was not flushed.\n", argv0);
    bx_dbg_exit(1);
    }

  // if Q already has MAX elements in it
  if (batch_dma.Qsize >= BX_BATCH_DMA_BUFSIZE) {
    fprintf(stderr, "%s: DMA batch buffer overrun.\n", argv0);
    bx_dbg_exit(1);
    }

  batch_dma.Qsize++;
  batch_dma.Q[batch_dma.Qsize-1].addr   = addr;
  batch_dma.Q[batch_dma.Qsize-1].len    = len;
  batch_dma.Q[batch_dma.Qsize-1].what   = what;
  batch_dma.Q[batch_dma.Qsize-1].val    = val;
  batch_dma.Q[batch_dma.Qsize-1].icount = bpoint_state.instr_count;

  // if Q is full, post events (and flush)
  if (batch_dma.Qsize >= batch_dma.this_many)
    bx_dbg_post_dma_reports();
}


  void
bx_dbg_irq_report(unsigned irq)
{
  if (bx_dbg_event_reports) {
    fprintf(stderr, "event icount=%u IRQ irq=%u\n",
      (unsigned) bpoint_state.instr_count, irq);
    }
}

  void
bx_dbg_a20_report(unsigned val)
{
  if (bx_dbg_event_reports) {
    fprintf(stderr, "event icount=%u A20 val=%u\n",
      (unsigned) bpoint_state.instr_count, val);
    }
}

  void
bx_dbg_io_report(Bit32u addr, unsigned size, unsigned op, Bit32u val)
{
  if (bx_dbg_event_reports) {
    fprintf(stderr, "event icount=%u IO addr=0x%x size=%u op=%s val=0x%x\n",
      (unsigned) bpoint_state.instr_count,
      (unsigned) addr,
      size,
      (op==BX_READ) ? "read" : "write",
      (unsigned) val);
    }
}

  void
bx_dbg_ucmem_report(Bit32u addr, unsigned size, unsigned op, Bit32u val)
{
  if (bx_dbg_event_reports) {
    fprintf(stderr, "event icount=%u UCmem addr=0x%x size=%u op=%s val=0x%x\n",
      (unsigned) bpoint_state.instr_count,
      (unsigned) addr,
      size,
      (op==BX_READ) ? "read" : "write",
      (unsigned) val);
    }
}

  void
bx_dbg_set_cpu_command(void)
{
  FILE *fp;
  int   reti;
  char *rets;
  Boolean retb;

  bx_dbg_cpu_t cpu;

  fp = bx_infile_stack[bx_infile_stack_index].fp;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "eax:0x%lx", &cpu.eax);
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "ebx:0x%lx", &cpu.ebx);
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "ecx:0x%lx", &cpu.ecx);
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "edx:0x%lx", &cpu.edx);
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "ebp:0x%lx", &cpu.ebp);
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "esi:0x%lx", &cpu.esi);
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "edi:0x%lx", &cpu.edi);
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "esp:0x%lx", &cpu.esp);
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "eflags:0x%lx", &cpu.eflags);
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "eip:0x%lx", &cpu.eip);
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "cs:s=0x%lx, dl=0x%lx, dh=0x%lx, valid=%lu",
                &cpu.cs.sel, &cpu.cs.des_l, &cpu.cs.des_h, &cpu.cs.valid);
  if (reti != 4) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "ss:s=0x%lx, dl=0x%lx, dh=0x%lx, valid=%lu",
                &cpu.ss.sel, &cpu.ss.des_l, &cpu.ss.des_h, &cpu.ss.valid);
  if (reti != 4) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "ds:s=0x%lx, dl=0x%lx, dh=0x%lx, valid=%lu",
                &cpu.ds.sel, &cpu.ds.des_l, &cpu.ds.des_h, &cpu.ds.valid);
  if (reti != 4) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "es:s=0x%lx, dl=0x%lx, dh=0x%lx, valid=%lu",
                &cpu.es.sel, &cpu.es.des_l, &cpu.es.des_h, &cpu.es.valid);
  if (reti != 4) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "fs:s=0x%lx, dl=0x%lx, dh=0x%lx, valid=%lu",
                &cpu.fs.sel, &cpu.fs.des_l, &cpu.fs.des_h, &cpu.fs.valid);
  if (reti != 4) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "gs:s=0x%lx, dl=0x%lx, dh=0x%lx, valid=%lu",
                &cpu.gs.sel, &cpu.gs.des_l, &cpu.gs.des_h, &cpu.gs.valid);
  if (reti != 4) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "ldtr:s=0x%lx, dl=0x%lx, dh=0x%lx, valid=%lu",
                &cpu.ldtr.sel, &cpu.ldtr.des_l, &cpu.ldtr.des_h, &cpu.ldtr.valid);
  if (reti != 4) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "tr:s=0x%lx, dl=0x%lx, dh=0x%lx, valid=%lu",
                &cpu.tr.sel, &cpu.tr.des_l, &cpu.tr.des_h, &cpu.tr.valid);
  if (reti != 4) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "gdtr:base=0x%lx, limit=0x%lx",
                &cpu.gdtr.base, &cpu.gdtr.limit);
  if (reti != 2) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "idtr:base=0x%lx, limit=0x%lx",
                &cpu.idtr.base, &cpu.idtr.limit);
  if (reti != 2) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "dr0:0x%lx", &cpu.dr0);
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "dr1:0x%lx", &cpu.dr1);
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "dr2:0x%lx", &cpu.dr2);
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "dr3:0x%lx", &cpu.dr3);
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "dr4:0x%lx", &cpu.dr4);
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "dr5:0x%lx", &cpu.dr5);
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "dr6:0x%lx", &cpu.dr6);
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "dr7:0x%lx", &cpu.dr7);
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "tr3:0x%lx", &cpu.tr3);
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "tr4:0x%lx", &cpu.tr4);
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "tr5:0x%lx", &cpu.tr5);
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "tr6:0x%lx", &cpu.tr6);
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "tr7:0x%lx", &cpu.tr7);
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "cr0:0x%lx", &cpu.cr0);
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "cr1:0x%lx", &cpu.cr1);
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "cr2:0x%lx", &cpu.cr2);
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "cr3:0x%lx", &cpu.cr3);
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "cr4:0x%lx", &cpu.cr4);
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "inhibit_int:%u", &cpu.inhibit_int);
  if (reti != 1) goto scanf_error;

  rets = fgets(tmp_buf, 512, fp); if (!rets) goto eof_error;
  reti = sscanf(tmp_buf, "done");
  if (reti != 0) goto scanf_error;

  retb = bx_dbg_set_cpu(&cpu);
  if (retb == 0)
    fprintf(stderr, "Error: set_cpu encountered error\n");
  else
    fprintf(stderr, "OK\n");
  return;

eof_error:
  fprintf(stderr, "Error: EOF encountered in set_cpu input stream\n");
  return;

scanf_error:
  fprintf(stderr, "Error: scanf returned error in set_cpu input stream\n");
  return;
}

  void
bx_dbg_disassemble_command(Bit32u addr1, Bit32u addr2)
{
  Boolean paddr_valid;
  Bit32u  paddr;
  unsigned ilen;

  do {
    dbg_xlate_linear2phy(addr1, &paddr, &paddr_valid);

    if (paddr_valid) {
#if BX_DISASM

      bx_dbg_fetch_mem(paddr, 16, bx_disasm_ibuf);
      ilen = bx_disassemble.disasm(bx_dbg_disassemble_size==32,
                                   bx_disasm_ibuf, bx_disasm_tbuf);
   
      fprintf(stderr, "%08x: ", (unsigned) addr1);
      for (unsigned j=0; j<ilen; j++)
        fprintf(stderr, "%02x", (unsigned) bx_disasm_ibuf[j]);
      fprintf(stderr, ": %s\n", bx_disasm_tbuf);
#endif  // #if BX_DISASM
      }
    else {
      fprintf(stderr, "??? (physical address not available)\n");
      ilen = 0; // keep compiler happy
      addr1 = addr2; // bail out
      }
      addr1 += ilen;
    } while (addr1 < addr2);
}

  void
bx_dbg_instrument_command(char *comm)
{
#if BX_INSTRUMENTATION
  if ( !strcmp(comm, "start") ) {
    bx_instr_start();
    }
  else if ( !strcmp(comm, "stop") ) {
    bx_instr_stop();
    }
  else if ( !strcmp(comm, "reset") ) {
    bx_instr_reset();
    }
  else if ( !strcmp(comm, "print") ) {
    bx_instr_print();
    }
  else {
    fprintf(stderr, "Error: command instrument %s not implemented.\n", comm);
    bx_dbg_exit(1);
    }
#else
    UNUSED(comm);

    fprintf(stderr, "Error: instrumentation not enabled.\n");
#endif
}

  void
bx_dbg_loader_command(char *path_quoted)
{
  size_t len;

  // skip beginning double quote
  if (path_quoted[0] == '"')
    path_quoted++;

  // null out ending quote
  len = strlen(path_quoted);
  if (path_quoted[len - 1] == '"')
    path_quoted[len - 1] = '\0';

#if BX_USE_LOADER
  bx_dbg_loader(path_quoted);
#else
  fprintf(stderr, "Error: loader not implemented.\n");
#endif
}

  void
bx_dbg_doit_command(void)
{
  // generic command to add temporary hacks to
  // for debugging purposes
}

  void
bx_dbg_crc_command(Bit32u addr1, Bit32u addr2)
{
  unsigned long crc;
  unsigned char *databuf;
  unsigned len;

  if (addr1 >= addr2) {
    fprintf(stderr, "Error: crc: invalid range.\n");
    return;
    }

  // get buf
  len = (addr2 - addr1) + 1;
  databuf = (unsigned char *) malloc(len);
  if (databuf == NULL) {
    fprintf(stderr, "Error: crc: out of memory.\n");
    return;
    }
  bx_dbg_fetch_mem(addr1, len, databuf);


  crc = crc32(databuf, len);
  free(databuf);

  fprintf(stderr, "0x%lx\n", crc);
}

  void
bx_dbg_info_dirty(void)
{
  unsigned i;

  for (i=0; i<sizeof(bx_dbg_dirty_pages); i++) {
    if (bx_dbg_dirty_pages[i]) {
      fprintf(stderr, "0x%x\n", i);
      bx_dbg_dirty_pages[i] = 0; // reset to clean
      }
    }
}
