//
// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
// 
//     EMail:    bochs@world.std.com



#include "bochs.h"




  // constructor
bx_pc_system_c::bx_pc_system_c(void)
{
  num_timers = 0;
  min_timer_index = 0;
  num_cpu_ticks_left = 1;
  m_ips = 0.0L;

  for (unsigned int i=0; i < 8; i++) {
    DRQ[i] = 0;
    DACK[i] = 0;
    }
  TC = 0;
  HRQ = 0;
  HLDA = 0;

  enable_a20 = 1;
 
#if BX_CPU < 2
  a20_mask   =    0xfffff;
#elif BX_CPU == 2
  a20_mask   =   0xffffff;
#else /* 386+ */
  a20_mask   = 0xffffffff;
#endif

}

  void
bx_pc_system_c::init_ips(Bit32u ips)
{
  // parameter 'ips' is the processor speed in Instructions-Per-Second
  m_ips = double(ips) / 1000000.0L;
fprintf(stderr, "# ips = %u\n", (unsigned) ips);
}

  inline void
bx_pc_system_c::new_min_timer(void)
{
  /* convenience function to determine the timer with the lowest
   * remaining time left; the next to go off
   */
  unsigned i;
  Bit32u   min_time;
  Boolean  found;

  min_time = (Bit32u) -1;
  found = 0;
  for (i=0; i<num_timers; i++) {
    if (timer[i].active && (found==0 || timer[i].remaining<min_time)) {
      found = 1;
      min_timer_index = i;
      min_time = timer[i].remaining;
      }
    }

  if (found == 0)
    bx_panic("new_min_timer(): no more active timers\n");
}


  void
bx_pc_system_c::timer_handler(void)
{
  Bit32u previous_min;
  unsigned i;


  /* subtract last timer interval from all remaining times */
  previous_min = timer[min_timer_index].remaining;

  for (i=0; i < num_timers; i++) {
    timer[i].triggered = 0;
    if (timer[i].active) {
      timer[i].remaining -= previous_min;
      if (timer[i].remaining == 0) {
        timer[i].triggered = 1;
        /* reset remaining period for triggered timer */
        timer[i].remaining = timer[i].period;

        /* if triggered timer is one-shot, deactive */
        if (timer[i].continuous==0)
          timer[i].active = 0;

        }
      }
    }

  new_min_timer();

  for (i=0; i < num_timers; i++) {
    /* call requested timer function.  It may request a different
     * timer period or deactivate, all cases handled below
     */
    if (timer[i].triggered && timer[i].funct!=NULL)
      timer[i].funct(timer[i].this_ptr);
    }

  new_min_timer();

  num_cpu_ticks_left = 1 +
      (Bit32u) ( double(timer[min_timer_index].remaining) * m_ips );

}


  /* REGISTER_TIMER()
   *   funct:      a function to be called after useconds
   *   useconds:   the interval between calls to funct
   *   continuous: 0=1-shot invokation, 1=continuous calls
   *   active:     0=initially inactive, 1=intially active
   */
#if 0
  int
bx_pc_system_c::register_timer( void (*funct)(void), Bit32u useconds,
  Boolean continuous, Boolean active)
{
  if (num_timers >= BX_MAX_TIMERS) {
    bx_panic("register_timer: too many registered timers.");
    }

  num_timers++;
  timer[num_timers - 1].period    = useconds;
  timer[num_timers - 1].remaining = useconds;
  timer[num_timers - 1].active    = active;
  timer[num_timers - 1].funct     = funct;
  timer[num_timers - 1].continuous = continuous;

  /* return timer id */
  return(num_timers - 1);
}
#endif

  int
bx_pc_system_c::register_timer( void *this_ptr, void (*funct)(void *), Bit32u useconds,
  Boolean continuous, Boolean active)
{
  if (num_timers >= BX_MAX_TIMERS) {
    bx_panic("register_timer: too many registered timers.");
    }

  num_timers++;
  timer[num_timers - 1].period    = useconds;
  timer[num_timers - 1].remaining = useconds;
  timer[num_timers - 1].active    = active;
  timer[num_timers - 1].funct     = funct;
  timer[num_timers - 1].continuous = continuous;
  timer[num_timers - 1].this_ptr   = this_ptr;

  /* return timer id */
  return(num_timers - 1);
}


  void
bx_pc_system_c::start_timers(void)
{
  if (!num_timers)
    bx_panic("start_timers(): no timers found\n");

  new_min_timer();

  num_cpu_ticks_left = 1 +
      (Bit32u) ( double(timer[min_timer_index].remaining) * m_ips );

bx_printf("start_timers(): timer %u intially %u ticks\n",
  (unsigned) min_timer_index, (unsigned) num_cpu_ticks_left);
}

  Bit32u
bx_pc_system_c::get_timer(unsigned index)
{
  Bit32u u_elapsed, u_remainder, min_u_remainder;

  if (index >= num_timers)
    bx_panic("bx_get_timer(): bad timer index %d\n", index);

  /* 1st get the useconds left for the next timer to go off */
  min_u_remainder = Bit32u( double(num_cpu_ticks_left) * m_ips );

  // clamp between 1 and remaining
  if (min_u_remainder < 1)
    min_u_remainder = 1;

  if (min_u_remainder > timer[min_timer_index].remaining)
    min_u_remainder = timer[min_timer_index].remaining;

  /* if requested timer is next to go off */
  if (index == min_timer_index) {
    return(min_u_remainder);
    }

  /* else calculate elapsed time from timer-at-bat, and subtract */
  else {
    u_elapsed = timer[min_timer_index].remaining - min_u_remainder;
    u_remainder = timer[index].remaining - u_elapsed;
    return(u_remainder);
    }
}

  void
bx_pc_system_c::activate_timer( unsigned timer_index,
    Bit32u useconds, Boolean continuous )
{
  Bit32u spent_time, current_remaining;
  unsigned i;

  if (timer_index >= num_timers)
    bx_panic("activate_timer(): bad timer index given\n");

  // set timer continuity to new value (1=continuous, 0=one-shot)
  timer[timer_index].continuous = continuous;

  current_remaining = Bit32u( double(num_cpu_ticks_left) * m_ips );
  if (current_remaining < 1)
    current_remaining = 1;

  if (current_remaining > timer[min_timer_index].remaining)
    current_remaining = timer[min_timer_index].remaining;

  spent_time = (timer[min_timer_index].remaining - current_remaining);
  for (i=0; i<num_timers; i++) {
    if (timer[i].active)
      timer[i].remaining -= spent_time;
    }

  /* if useconds = 0, use default stored in period field
   * else set new period from useconds */
  if (useconds==0)
    useconds = timer[timer_index].period;
  else
    timer[timer_index].period = useconds;

  timer[timer_index].active = 1;
  timer[timer_index].remaining = timer[timer_index].period;

  new_min_timer();

  num_cpu_ticks_left = 1 +
      (Bit32u) ( double(timer[min_timer_index].remaining) * m_ips );
}

  void
bx_pc_system_c::deactivate_timer( unsigned timer_index )
{
  Bit32u spent_time, current_remaining;
  unsigned i;

  if (timer_index >= num_timers)
    bx_panic("deactivate_timer(): bad timer index given\n");

  current_remaining = Bit32u( double(num_cpu_ticks_left) * m_ips );
  if (current_remaining < 1)
    current_remaining = 1;

  if (current_remaining > timer[min_timer_index].remaining)
    current_remaining = timer[min_timer_index].remaining;

  spent_time = (timer[min_timer_index].remaining - current_remaining);
  for (i=0; i<num_timers; i++) {
    if (timer[i].active)
      timer[i].remaining -= spent_time;
    }

  timer[timer_index].active = 0;
  /* if currently scheduled timer is requested timer to deactivate */
  if (min_timer_index == timer_index) {
    new_min_timer();
    }

  num_cpu_ticks_left = 1 +
      (Bit32u) ( double(timer[min_timer_index].remaining) * m_ips );
}

  void
bx_pc_system_c::raise_HLDA(void)
{
  HLDA = 1;
  bx_devices.raise_hlda();
  HLDA = 0;
}

  void
bx_pc_system_c::set_DRQ(unsigned channel, Boolean val)
{
  if (channel > 7)
    bx_panic("set_DRQ() channel > 7\n");
  DRQ[channel] = val;
  bx_devices.drq(channel, val);
}

  void
bx_pc_system_c::set_HRQ(Boolean val)
{
  HRQ = val;
  if (val)
    bx_cpu.async_event = 1;
  else
    HLDA = 0; // ??? needed?
}

  void
bx_pc_system_c::set_TC(Boolean val)
{
  TC = val;
}

  void
bx_pc_system_c::set_DACK(unsigned channel, Boolean val)
{
  DACK[channel] = val;
}


  void
bx_pc_system_c::dma_write8(Bit32u phy_addr, unsigned channel)
{
  // DMA controlled xfer of byte from I/O to Memory

  Bit8u data_byte;

  UNUSED(channel);
  bx_devices.dma_write8(channel, &data_byte);
  bx_mem.write_physical(phy_addr, 1, &data_byte);

  BX_DBG_DMA_REPORT(phy_addr, 1, BX_WRITE, data_byte);
}


  void
bx_pc_system_c::dma_read8(Bit32u phy_addr, unsigned channel)
{
  // DMA controlled xfer of byte from Memory to I/O

  Bit8u data_byte;

  UNUSED(channel);
  bx_mem.read_physical(phy_addr, 1, &data_byte);
  bx_devices.dma_read8(channel, &data_byte);

  BX_DBG_DMA_REPORT(phy_addr, 1, BX_READ, data_byte);
}

  void
bx_pc_system_c::set_INTR(Boolean value)
{
  bx_cpu.set_INTR(value);
}

//
// Read a byte of data from the IO memory address space
//

  Bit32u
bx_pc_system_c::inp(Bit16u addr, unsigned io_len)
{
  Bit32u ret;

  ret = bx_devices.inp(addr, io_len);

  BX_DBG_IO_REPORT(addr, io_len, BX_READ, ret);

  if (bx_dbg.io)
    bx_printf("IO: inp(%04x) = %02x\n", (unsigned) addr, (unsigned) ret);

  return( ret );
}


//
// Write a byte of data to the IO memory address space.
//

  void
bx_pc_system_c::outp(Bit16u addr, Bit32u value, unsigned io_len)
{
  bx_devices.outp(addr, value, io_len);

  BX_DBG_IO_REPORT(addr, io_len, BX_WRITE, value);

  if (bx_dbg.io)
    bx_printf("IO: outp(%04x) = %02x\n", (unsigned) addr, (unsigned) value);
}


  void
bx_pc_system_c::set_enable_a20(Bit8u value)
{
#if BX_CPU < 2
    bx_panic("set_enable_a20() called: 8086 emulation\n");
#else

#if BX_SUPPORT_A20
  if (value) {
    enable_a20 = 1;
#if BX_CPU == 2
    a20_mask   = 0xffffff;   /* 286: enable all 24 address lines */
#else /* 386+ */
    a20_mask   = 0xffffffff; /* 386: enable all 32 address lines */
#endif
    }
  else {
    enable_a20 = 0;
    a20_mask   = 0xffefffff;   /* mask off A20 address line */
    }

  BX_DBG_A20_REPORT(value);

  if (bx_dbg.a20)
    bx_printf("A20: set() = %u\n", (unsigned) enable_a20);
#else
  bx_printf("set_enable_a20: ignoring: SUPPORT_A20 = 0\n");
#endif  // #if BX_SUPPORT_A20

#endif
}

  Boolean
bx_pc_system_c::get_enable_a20(void)
{
#if BX_SUPPORT_A20
  if (bx_dbg.a20)
    bx_printf("A20: get() = %u\n", (unsigned) enable_a20);

  if (enable_a20) return(1);
  else return(0);
#else
  bx_printf("get_enable_a20: ignoring: SUPPORT_A20 = 0\n");
  return(1);
#endif  // #if BX_SUPPORT_A20
}

  int
bx_pc_system_c::ResetSignal( PCS_OP operation )
{
  UNUSED( operation );
  // Reset the processor.
 
  fprintf(stderr, "# bx_pc_system_c::ResetSignal() called\n");
  ::exit(1);
 
  return 0;
}
 
 
  PCS_OP
bx_pc_system_c::GetA20Signal( void )
{
  return( bx_pc_system.get_enable_a20() ? PCS_CLEAR : PCS_SET );
}
 
 
  int
bx_pc_system_c::A20Signal( PCS_OP operation )
{
  // Set or clear the A20 mask.
 
  switch ( operation ) {
    case PCS_CLEAR:
      //fprintf(stderr, "# bx_pc_system_c::A20Signal(PCS_CLEAR) called\n");
      bx_pc_system.set_enable_a20(1);
      break;
    case PCS_SET:
      //fprintf(stderr, "# bx_pc_system_c::A20Signal(PCS_SET) called\n");
      bx_pc_system.set_enable_a20(0);
      break;
    case PCS_TOGGLE:
      fprintf(stderr, "# bx_pc_system_c::A20Signal(PCS_TOGGLE) called\n");
      ::exit(1);
      break;
    }
  return 0;
}

  Bit8u
bx_pc_system_c::IAC(void)
{
  return( bx_devices.pic->IAC() );
}

  void
bx_pc_system_c::exit(void)
{
  if (bx_devices.hard_drive)
    bx_devices.hard_drive->close_harddrive();
  bx_gui.exit();
}
